package NelinearneStrukture;

/**
 * Klasa koja implementira ponašanje binarnog stabla. 
 * Ovo je dinamička razgranata struktura kod koje svaki element ima
 * najviše jednog prethodnika i najviše dva sledbenika. Elementi mogu da se 
 * dodaju bilo gde u strukturu i bilo koji element može da se izbaci. 
 * Redosled ubacivanja i izbacivanja nije eksplicitno definisan. 
 * 
 * @author Kostandin Dimitrijević, Strukture podataka i algoritmi, FON, 2010
 */
public class BinarnoStablo {
	// Pokazivač na koreni (prvi) čvor u stablu.
	CvorStabla koren;
	
	public BinarnoStablo(){
		koren = null;
	}
	
	/**
	 * Proverava da li je stablo prazno. Stablo je prazno kada ne postoji 
	 * pokazivač na koren.
	 * @return true ako je stablo prazno. u suprotnom false
	 */
	public boolean praznoStablo(){
		return koren == null;
	}
	
	/**
	 * Primer implementacije prefiksnog prolaza kroz stablo. 
	 * U ovom primeru se vrši ispisivanje elemenata u prefiksnom redosledu. 
	 * Stablo se obilazi u redosledu koren - levo - desno. 
	 * Levo i desno podstablo se obilaze na isti način.
	 * @param cvor Pokazivač na koren stabla koje se obilazi.
	 */
	public void prefiksniProlaz(CvorStabla cvor){
		if (cvor == null) return;

		// ispiši koren (trenutni element)
		System.out.println(cvor.podatak);
		// obiđi prefiksno levo podstablo
		prefiksniProlaz(cvor.levo);
		// obiđi prefiksno desno podstablo
		prefiksniProlaz(cvor.desno);
	}
	
	/**
	 * Primer implementacije infiksnog prolaza kroz stablo. 
	 * U ovom primeru se vrši ispisivanje elemenata u infiksnom redosledu. 
	 * Stablo se obilazi u redosledu levo - koren - desno. 
	 * Levo i desno podstablo se obilaze na isti način.
	 * @param cvor Pokazivač na koren stabla koje se obilazi.
	 */
	public void infiksniProlaz(CvorStabla cvor){
		if (cvor == null) return;

		// obiđi infiksno levo podstablo
		infiksniProlaz(cvor.levo);
		// ispiši koren (trenutni element)
		System.out.println(cvor.podatak);
		// obiđi infiksno desno podstablo
		infiksniProlaz(cvor.desno);
	}

	/**
	 * Primer implementacije postfiksnog prolaza kroz stablo. 
	 * U ovom primeru se vrši ispisivanje elemenata u postfiksnom redosledu. 
	 * Stablo se obilazi u redosledu levo - desno - koren. 
	 * Levo i desno podstablo se obilaze na isti način.
	 * @param cvor Pokazivač na koren stabla koje se obilazi.
	 */
	public void postfiksniProlaz(CvorStabla cvor){
		if (cvor == null) return;

		// obiđi postfiksno levo podstablo
		postfiksniProlaz(cvor.levo);
		// obiđi postfiksno desno podstablo
		postfiksniProlaz(cvor.desno);
		// ispiši koren (trenutni element)
		System.out.println(cvor.podatak);
	}

	/**
	 * Metoda za prebrojavanje čvorova u stablu. 
	 * Prolazi kroz celo stablo na prefiksni način i pri obilasku svakog čvora
	 * dodaje 1 na ukupan broj čvorova.
	 * @param cvor Pokazivač na koren stabla za koje tražimo broj elemenata
	 * @return Broj čvorova u stablu.
	 */
	public int brojElemenata(CvorStabla cvor){
		if (cvor == null) return 0;
		// broj čvorova u stablu je jednak 
		// koren (1) + broj elemenata u levom podstablu + broj elemenata u desnom podstablu
		// prvi nacin
//		int br = 1; // računaj trenutni čvor
		// dodaj broj elemenata u levom podstablu
//		br += brojElemenata(cvor.levo);
		// dodaj broj elemenata u desnom podstablu
//		br += brojElemenata(cvor.desno);
//		return br;
		
		// drugi nacin
		return 1 + brojElemenata(cvor.levo) + brojElemenata(cvor.desno);
	}

	/**
	 * Metoda za izračunavanje zbira elemenata čvorova stabla. 
	 * Prolazi kroz celo stablo na prefiksni način i pri obilasku svakog čvora
	 * dodaje njegovu vrednost na ukupan zbir.
	 * @param cvor Pokazivač na koren stabla za koje tražimo zbir elemenata
	 * @return Zbir elemenata čvorova u stablu.
	 */
	public int zbirElemenata(CvorStabla cvor){
		if (cvor == null) return 0;
		// zbir elemenata čvorova u stablu je jednak 
		// vrednost u korenu + zbir elemenata u levom podstablu + zbir elemenata u desnom podstablu
		return cvor.podatak + zbirElemenata(cvor.levo) + zbirElemenata(cvor.desno);
	}
	
	/**
	 * Metoda za pronalazak maksimalne vrednosti u stablu. 
	 * Rekurzivno pronalazi maksimalni element u levom podstablu i 
	 * maksimalni element u desnom podstablu.
	 * Nakon toga upoređuje te dve vrednosti sa vrednošću u trenutnom čvoru
	 * i vraća najveću vrednost od te tri vrednosti.
	 * @param cvor Pokazivač na koren stabla za koje tražimo maksimalni element
	 * @return Vrednost maksimalnog elementa u stablu.
	 */
	public int maxElement(CvorStabla cvor){
		if (cvor == null) return Integer.MIN_VALUE;
		
		// prvi nacin
		int max = cvor.podatak;
		// nađi maksimalnu vrednost u levom podstablu
		int maxl = maxElement(cvor.levo);
		// nađi maksimalnu vrednost u desnom podstablu
		int maxd = maxElement(cvor.desno);
		
		// upoređivanje vrednosti. max će uvek uzimati veću vrednost
		if (max < maxl)
			max = maxl;
		
		if (max < maxd)
			max = maxd;
		
		return max;
		
		// drugi nacin
		//return Math.max(Math.max(cvor.podatak, maxElement(cvor.levo)), maxElement(cvor.desno));
	}

	public int minElement(CvorStabla cvor){
		if (cvor == null) return Integer.MAX_VALUE;
		
		// prvi nacin
		int min = cvor.podatak;
		int minl = minElement(cvor.levo);
		int mind = minElement(cvor.desno);
		
		if (min > minl)
			min = minl;
		
		if (min > mind)
			min = mind;
		
		return min;
	}

	int visina(CvorStabla cvor){
		if (cvor == null || (cvor.levo == null && cvor.desno == null))
			return 0;
		
		return 1 + Math.max(visina(cvor.levo), visina(cvor.desno));
	}
	
	CvorStabla pronadji(CvorStabla tekuci, int podatak){
		if (tekuci == null || tekuci.podatak == podatak)
			return tekuci;

		CvorStabla l = pronadji(tekuci.levo, podatak);
		if (l != null)
			return l;
		
		// prvi nacin
		CvorStabla d = pronadji(tekuci.desno, podatak);
		if (d != null)
			return d;

		return null;
		
		// drugi nacin
		//return pronadji(tekuci.desno, podatak);
	}
	
	
	boolean daLiJeBst(CvorStabla cvor){
		if (cvor == null)
			return true;
		
		if (cvor.podatak > maxElement(cvor.levo) && 
				cvor.podatak < minElement(cvor.desno)) {
			
			return daLiJeBst(cvor.levo) && daLiJeBst(cvor.desno);
		}
		
		return false;
	}

	boolean daLiJeAvl(CvorStabla cvor){
		if (cvor == null)
			return true;
		
		if (daLiJeBst(cvor)){
			int rv = visina(cvor.levo) - visina(cvor.desno);

//			if (Math.abs(rv) < 2)
			if (rv > -2 && rv < 2)
				return daLiJeAvl(cvor.levo) && daLiJeAvl(cvor.desno);
		}
		return false;
	}
	
	boolean daLiJeHeap(CvorStabla cvor){
		if (cvor == null)
			return true;
		
		if ((cvor.levo == null || cvor.podatak > cvor.levo.podatak) &&
				cvor.desno == null || cvor.podatak > cvor.desno.podatak)
		
//		if (cvor.podatak == maxElement(cvor))
			return daLiJeHeap(cvor.levo) && daLiJeHeap(cvor.desno);

		return false;
	}
}
