package NelinearneStrukture;

/**
 * Klasa koja implementira BST stablo.
 * BST stablo je binarno stablo optimizovano za pretragu 
 * (Binarno Stablo Traženja ili Binary Search Tree). Kod ovog stabla su
 * elementi u njemu raspoređeni tako da za svaki vor važi da su vrednosti
 * u čvorovima njegovog levog podstabla manje od vrednosti u tom čvoru, 
 * i vrednosti u čvorovima njegovog desnog podstabla veće od vrednosti u tom čvoru.
 * Pošto je ovo binarno stablo, ova klasa nasleđuje klasu BinarnoStablo.
 * 
 * @author Kostandin Dimitrijević, Strukture podataka i algoritmi, FON, 2010
 */
public class BstStablo extends BinarnoStablo {

	/**
	 * Glavna (javna) metoda za dodavanje novog elementa u stablo. 
	 * Poziva privatnu metodu sa odgovarajućim parametrima.
	 * @param podatak Podatak koji se ubacuje u stablo.
	 */
	public void ubaciUStablo(int podatak){
		ubaci(koren, podatak);
	}
	
	/**
	 * Metoda koja implementira algoritam za ubacivanje novog elementa u BST stablo.
	 * Nov element se uvek ubacuje kao novi list u stablu, a mesto mu se određuje
	 * upoređivanjem vrednosti sa vrednošću u trenutnom čvoru. 
	 * 
	 * @param tekuci Pokazivač na tekući čvor u stablu. Novi podatak ubacujemo u 
	 * njegovo levo ili desno podstablo
	 * @param podatak Element koji se ubacuje u stablo
	 */
	private void ubaci(CvorStabla tekuci, int podatak){
		if (praznoStablo())
			koren = new CvorStabla(podatak);
		else {
			// proveravamo da li je vrednost podatka koju ubacujemo
			// veća ili manja od vrednosti u tekućem čvoru
			if (podatak < tekuci.podatak){
				// ako je vrednost manja onda se novi element ubacuje u levo podstablo
				if (tekuci.levo != null)
					// ubacivanje u levo podstablo
					ubaci(tekuci.levo, podatak);
				else 
					// ako ne postoji levo dete, onda se novi čvor ubacuje levo od tekućeg
					tekuci.levo = new CvorStabla(podatak);
			} else if (podatak > tekuci.podatak){
				// ako je vrednost veća onda se novi element ubacuje u desno podstablo
				if (tekuci.desno != null)
					// ubacivanje u desno podstablo
					ubaci(tekuci.desno, podatak);
				else
					// ako ne postoji desno dete, onda se novi čvor ubacuje desno od tekućeg
					tekuci.desno = new CvorStabla(podatak);
			}
			// ako je podatak jednak podatku u tekućem čvoru ne vrši se nikakva operacija
			// jer stablo ne može da ima duple vrednosti
		}
	}
	
	/**
	 * Pretraživanje BST stabla. Koristi se organizacija stabla kako bi se 
	 * što brže došlo do traženog elementa
	 * 
	 * @param tekuci Pokazivač na tekuči element u stablu
	 * @param podatak Podatak koji se traži
	 * @return Pokazivač na čvor koji ima traženu vrednost. Ako čvor ne postoji vraća null.
	 */
	CvorStabla pronadji(CvorStabla tekuci, int podatak){
		// ako smo stigli do "kraja" stabla (prvi uslov) ili smo pronašli čvor
		if (tekuci == null || tekuci.podatak == podatak)
			return tekuci;
		
		if (podatak < tekuci.podatak)
			return pronadji(tekuci.levo, podatak);
		else 
			return pronadji(tekuci.desno, podatak);
	}
	
	/**
	 * Pronalazak roditeljskog čvora za trenutni čvor
	 * @param tekuci Pokazivač na tekući čvor
	 * @param podatak Podatak za koji se traži roditeljski čvor
	 * @return Pokazivač na roditeljski čvor. Ako se podatak ne nalazi u stablu
	 * ili ako se podatak nalazi u korenu, vraća null.
	 */
	private CvorStabla nadjiRoditelja(CvorStabla tekuci, int podatak){
		if (tekuci == null || tekuci.podatak == podatak)
			return null;
	
		if (podatak < tekuci.podatak){
			// ako se podatak nalazi u levom detetu, onda je tekući njegov roditelj
			if (tekuci.levo != null && tekuci.levo.podatak == podatak)
				return tekuci;
			// probaj da pronađeš element u levom podstablu
			return nadjiRoditelja(tekuci.levo, podatak);
		} else {
			// ako se podatak nalazi u desnom detetu, onda je tekući njegov roditelj
			if (tekuci.desno != null && tekuci.desno.podatak == podatak)
				return tekuci;
			// probaj da pronađeš element u desnom podstablu
			return nadjiRoditelja(tekuci.desno, podatak);
		}
	}

	/**
	 * Ispisivanje putanje od korena do nekog čvora u stablu. 
	 * Koristi organizaciju BST stabla da bi se kretao pravom putanjom do čvora
	 * @param tekuci Pokazivač na tekući čvor na putanji 
	 * @param kraj Pokazivač na krajnji čvor na putanji
	 */
	void ispisiPutanju(CvorStabla tekuci, CvorStabla kraj){
		System.out.println(tekuci.podatak);
		
		if (tekuci == kraj) 
			return;
		// na osnovu vrednosti u čvorovima se određuje da li se dalje treba
		// kretati levo ili desno
		if (kraj.podatak < tekuci.podatak)
			ispisiPutanju(tekuci.levo, kraj);
		else
			ispisiPutanju(tekuci.desno, kraj);
	}

	/**
	 * Izračunavanje broja čvorova na putanji od korena do nekog čvora u stablu. 
	 * Koristi organizaciju BST stabla da bi se kretao pravom putanjom do čvora.
	 * Može se koristiti i za određivanje nivoa nekog čvora u stablu.
	 * @param tekuci Pokazivač na tekući čvor na putanji 
	 * @param kraj Pokazivač na krajnji čvor na putanji
	 * @return Broj elemenata na putanji od korena do datog čvora.
	 */
	int brElNaPutanji(CvorStabla tekuci, CvorStabla kraj){
		if (tekuci == kraj) 
			return 0;
		if (kraj.podatak < tekuci.podatak)
			return 1 + brElNaPutanji(tekuci.levo, kraj);
		else
			return 1 + brElNaPutanji(tekuci.desno, kraj);
	}
	
	
	private CvorStabla maxCvor(CvorStabla tekuci){
		while(tekuci.desno != null)
			tekuci = tekuci.desno;
		return tekuci;
	}

	private CvorStabla minCvor(CvorStabla tekuci){
		while(tekuci.levo != null)
			tekuci = tekuci.levo;
		return tekuci;
	}
	
	
	private void izbaciListPolulist(CvorStabla cvor){
		CvorStabla r = nadjiRoditelja(koren, cvor.podatak);
		CvorStabla dete = cvor.levo != null ? cvor.levo : cvor.desno;
		
		if (r == null)
			koren = dete;
		else {
			if (r.levo == cvor)
				r.levo = dete;
			else
				r.desno = dete;
		}
	}
	
	public void izbaci(int podatak){
		CvorStabla cvor = pronadji(koren, podatak);
		if (cvor == null) return;
		
		if (cvor.levo != null && cvor.desno != null){
			CvorStabla maxL = maxCvor(cvor.levo);
			izbaciListPolulist(maxL);
			cvor.podatak = maxL.podatak;
		} else 
			izbaciListPolulist(cvor);
	}




}
